# -*- coding: utf-8 -*-

"""Serve a takumi thrift service.

Usage:
    takumi serve [<gunicorn_args>...]
"""

import sys


def run(args):
    from ..runner import AppRunner
    from takumi_ext import ext
    sys.argv = ['takumi serve'] + args
    # Delegate to gunicorn
    runner = AppRunner()
    app_runner_ext = ext['app-runner']

    if app_runner_ext:
        runner_ext = app_runner_ext(runner)
        runner.cfg.set('when_ready', lambda x: runner_ext.on_start())
        runner.cfg.set('on_exit', lambda x: runner_ext.on_exit())
    runner.run()



# -*- coding: utf-8 -*-

"""Takumi command line toolkit.

Usage:
    takumi <command> [<args>...]
    takumi -h | --help

Options:
    -h, --help      Show this message and exit
    --version       Show version

Commands:
    help            Show help for command
    run             Run script
    test            Run tests
    serve           Serve the service using gunicorn
    deploy          Deploy the service to an environment
    shell           Start an IPython REPL
"""

import sys
import schema
from docopt import DocoptExit
from ._base import parse_args

__version__ = '0.1.0'

commands = 'help', 'run', 'test', 'serve', 'deploy', 'shell'


def run(command, args):
    if command == 'help':
        from .help import run as help
        cmd = help(args)
        return run(cmd, ['-h'])
    if command == 'serve':
        from .serve import run as serve
        return serve(args)
    if command == 'deploy':
        from .deploy import run as deploy
        return deploy(args)
    if command == 'shell':
        from .shell import run as shell
        return shell(args)
    raise DocoptExit('Command {!r} not supported\n'.format(command))


def main():
    args = parse_args(None, __doc__, sys.argv[1:], {
        '<command>': schema.And(
            str, lambda c: c in commands,
            error='`<command>` should be one of {}'.format(commands)),
        '<args>': list,
    }, version=__version__, options_first=True)

    if '' not in sys.path:
        # Insert current directory
        sys.path.insert(0, '')
    run(args['<command>'], args['<args>'])



# -*- coding: utf-8 -*-

import schema
from docopt import docopt, DocoptExit


def parse_args(cmd, doc, args, validator, **kwargs):
    argv = args if cmd is None else [cmd] + args
    args = docopt(doc, argv=argv, **kwargs)
    for item in ('--', '--help', cmd):
        args.pop(item, None)
    try:
        return schema.Schema(validator).validate(args)
    except schema.SchemaError as e:
        raise DocoptExit('{}\n'.format(e))



# -*- coding: utf-8 -*-

"""Start an IPython shell with predefined objects

Usage:
    takumi shell [-t HOST] [(-- <ipython_args>...)]
    takumi shell -h | --help

Options:
    -h, --help       Show this message and exit
    -t, --host HOST  Host the shell client connect to [default: localhost]
"""

import functools
import sys
import importlib

from ._base import parse_args


class _ClientWrapper(object):
    def __init__(self, pool, app):
        self._pool = pool
        self._app = app
        self._thrift = pool.thrift_module
        for func in pool.service.thrift_services:
            api = functools.partial(self.__call, func)
            if func in app.api_map:
                functools.wraps(app.api_map[func].func)(api)
            setattr(self, func, api)

    def __call(self, func, *args, **kwargs):
        with self._pool.client_ctx() as c:
            return getattr(c, func)(*args, **kwargs)


def _create_client_pool(host):
    from takumi_config import config
    if 'CLIENT_SETTINGS' not in config.settings:
        config.settings['CLIENT_SETTINGS'] = {}

    from takumi_client.pool import Pool
    client_name = '{}_shell'.format(config.app_name)

    mod, app = config.app.split(':')
    app = getattr(importlib.import_module(mod), app)
    pool = Pool(client_name, app.thrift_module, app.service_name,
                hosts=[(host, config.port)],
                client_version=config.version or '-')
    c = _ClientWrapper(pool, app)
    return c


def run(args):
    args = parse_args('shell', __doc__, args, {
        '<ipython_args>': list,
        '--host': str,
    })

    import gevent.monkey
    gevent.monkey.patch_all()
    from takumi_config import config

    client = _create_client_pool(args['--host'])
    banner = """Python {}

Interactive shell for service {}.
c         -> Client for invoking service api.
c._thrift -> Loaded thrift module.
config    -> Config entry.
""".format(sys.version.split('\n')[0].strip(), config.app_name)

    ns = {
        'config': config,
        'c': client
    }

    try:
        from IPython import start_ipython
        from traitlets.config.loader import Config
        c = Config(TerminalInteractiveShell={'banner1': banner})
        start_ipython(config=c, argv=args['<ipython_args>'], user_ns=ns)
    except ImportError:
        from code import interact
        interact(banner=banner, local=ns)
    finally:
        client._pool.close()



# -*- coding: utf-8 -*-

"""Deploy a Takumi app.

Usage:
    takumi deploy <target> [options] [(-- <ansible_args>...)]
    takumi deploy -- <ansible_args>...
    takumi deploy -h | --help

Options:
    -t, --tags TAGS  Only run tasks with these tags
    -p, --play PLAY  Specify a different playbook
    -h, --help       Show this message and exit

Example:

    Simple deploy:

        takumi deploy testing

    Deploy a subject:

        takumi deploy testing -t cron

    Specify other playbooks:

        takumi deploy testing -p system.yml
"""

import schema
from ._base import parse_args


def run(args):
    args = parse_args('deploy', __doc__, args, {
        '--tags': schema.Or(None, str),
        '--play': schema.Or(None, str),
        '<target>': schema.Or(None, lambda x: x != '--',
                              error='Invalid value of target'),
        '<ansible_args>': list
    })

    from ..deploy import start

    try:
        start(args)
    except Exception as e:
        exit('Fail to deploy: {}'.format(e))
